import { ImmerState, Patch, ImmerScope, Drafted, AnyObject, ImmerBaseState, AnyArray, AnyMap, AnySet, ProxyType } from "./internal";
/** Plugin utilities */
declare const plugins: {
    patches?: {
        generatePatches: typeof generatePatches;
        applyPatches: typeof applyPatches;
    };
    es5?: {
        willFinalizeES5: typeof willFinalizeES5;
        createES5Proxy: typeof createES5Proxy;
        markChangedES5: typeof markChangedES5;
    };
    mapset?: {
        proxyMap: typeof proxyMap;
        proxySet: typeof proxySet;
    };
};
declare type Plugins = typeof plugins;
export declare function getPlugin<K extends keyof Plugins>(pluginKey: K): Exclude<Plugins[K], undefined>;
export declare function loadPlugin<K extends keyof Plugins>(pluginKey: K, implementation: Plugins[K]): void;
/** ES5 Plugin */
interface ES5BaseState extends ImmerBaseState {
    finalizing: boolean;
    assigned: {
        [key: string]: any;
    };
    parent?: ImmerState;
    revoked: boolean;
}
export interface ES5ObjectState extends ES5BaseState {
    type: ProxyType.ES5Object;
    draft: Drafted<AnyObject, ES5ObjectState>;
    base: AnyObject;
    copy: AnyObject | null;
}
export interface ES5ArrayState extends ES5BaseState {
    type: ProxyType.ES5Array;
    draft: Drafted<AnyObject, ES5ArrayState>;
    base: AnyArray;
    copy: AnyArray | null;
}
export declare function willFinalizeES5(scope: ImmerScope, result: any, isReplaced: boolean): void;
export declare function createES5Proxy<T>(base: T, parent?: ImmerState): Drafted<T, ES5ObjectState | ES5ArrayState>;
export declare function markChangedES5(state: ImmerState): void;
/** Map / Set plugin */
export interface MapState extends ImmerBaseState {
    type: ProxyType.Map;
    copy: AnyMap | undefined;
    assigned: Map<any, boolean> | undefined;
    base: AnyMap;
    revoked: boolean;
    draft: Drafted<AnyMap, MapState>;
}
export interface SetState extends ImmerBaseState {
    type: ProxyType.Set;
    copy: AnySet | undefined;
    base: AnySet;
    drafts: Map<any, Drafted>;
    revoked: boolean;
    draft: Drafted<AnySet, SetState>;
}
export declare function proxyMap<T extends AnyMap>(target: T, parent?: ImmerState): T;
export declare function proxySet<T extends AnySet>(target: T, parent?: ImmerState): T;
/** Patches plugin */
export declare type PatchPath = (string | number)[];
export declare function generatePatches(state: ImmerState, basePath: PatchPath, patches: Patch[], inversePatches: Patch[]): void;
export declare function applyPatches<T>(draft: T, patches: Patch[]): T;
export {};
//# sourceMappingURL=plugins.d.ts.map